import csv
from datetime import datetime
from pathlib import Path

from capitalg.contstants import DATE_RATE_FORMAT
from capitalg.errors import InputValidationError


def load_rates(input_path: Path) -> dict:
    if input_path.exists() is False:
        return {}

    with open(input_path) as f:
        reader = csv.DictReader(f)
        rates = {}
        for i, daily_rates in enumerate(reader):
            try:
                date_key = datetime.fromisoformat(daily_rates['date']).strftime(DATE_RATE_FORMAT)
            except ValueError:
                raise InputValidationError(f'Error in rates file on line {i} - invalid date format')

            rates[date_key] = {
                asset.lower(): rate
                for asset, rate in daily_rates.items()
            }

    return rates



import csv
import logging
from datetime import datetime
from decimal import Decimal
from operator import itemgetter
from pathlib import Path
from typing import List, Optional

from capitalg.contstants import (
    DATE_INPUT_FORMAT,
    DATE_RATE_FORMAT,
    FIELD_ASSET_CODE,
    FIELD_BASE_CURRENCY,
    FIELD_FEE_CURRENCY,
    FIELD_DATE,
    FIELD_EXCHANGE,
    FIELD_FEE_UNIT,
    FIELD_FEE,
    FIELD_NOTE,
    FIELD_PRICE,
    FIELD_QTY,
    FIELD_RAW_ID,
    FIELD_TRANSACTION_TYPE,
    FIELD_TZ,
    TRANSACTION_BUY_LABEL,
    TRANSACTION_SELL_LABEL,
)
from capitalg.errors import InputValidationError
from capitalg.utils import convert_timezone

logger = logging.getLogger(__name__)

class TransactionLoader:

    def __init__(
        self,
        input_path: Path,
        output_path: Path,
        tax_currency: str,
        tax_year_cutoff: datetime,
        tax_timezone: Optional[str] = 'UTC',
        rates: Optional[dict] = None
    ):
        self.input_path = input_path
        self.output_path = output_path
        self.tax_currency = tax_currency
        self.tax_year_cutoff = tax_year_cutoff
        self.tax_timezone = tax_timezone
        self.transactions = []
        self.rates = rates
        self._load()

    @property
    def get_transactions(self):
        return self.transactions

    def _format_transaction(self, transaction: dict, transaction_date: datetime) -> dict:
        """ Converts numeric strings to decimal fields and format dates
        """
        fee = self._rebase_fee(transaction, transaction_date)
        return {
            FIELD_RAW_ID: transaction[FIELD_RAW_ID],
            FIELD_EXCHANGE: transaction[FIELD_EXCHANGE],
            FIELD_DATE: transaction_date.strftime(DATE_INPUT_FORMAT),
            FIELD_TZ: self.tax_timezone,
            FIELD_TRANSACTION_TYPE: transaction[FIELD_TRANSACTION_TYPE],
            FIELD_BASE_CURRENCY: self.tax_currency,
            FIELD_ASSET_CODE: transaction[FIELD_ASSET_CODE],
            FIELD_PRICE: Decimal(transaction[FIELD_PRICE]),
            FIELD_QTY: Decimal(transaction[FIELD_QTY]),
            FIELD_FEE_CURRENCY: self.tax_currency,
            FIELD_FEE: fee,
            FIELD_FEE_UNIT: round(fee / Decimal(transaction[FIELD_QTY]), 2) if Decimal(transaction[FIELD_QTY]) > 0 else Decimal(0),
            FIELD_NOTE: transaction[FIELD_NOTE],
        }

    def _load(self) -> List[dict]:
        with open(self.input_path) as f:
            reader = csv.DictReader(f)
            transactions = []
            for i, raw_transaction in enumerate(reader):
                
                self._validate_transaction(raw_transaction)
                transaction = self._standardize_transaction(raw_transaction)

                # Allow for zulu time, since it is a common export format
                # If zulu time is detected, we drop the Z and override the tz to be UTC
                uses_zulu_time = transaction[FIELD_DATE][-1].lower() == 'z'

                transaction_date = convert_timezone(
                    date_str=transaction[FIELD_DATE][:-1] if uses_zulu_time is True else transaction[FIELD_DATE],
                    source_tz='UTC' if uses_zulu_time is True else transaction[FIELD_TZ],
                    dest_tz=self.tax_timezone
                )

                if transaction_date >= self.tax_year_cutoff:
                    # Ignore transactions that have occurred after the tax year
                    continue

                try:
                    if self.tax_currency != transaction[FIELD_BASE_CURRENCY]:
                        transactions.extend(
                            self._rebase_transaction(transaction, transaction_date)
                        )
                    else:
                        transactions.append(
                            self._format_transaction(transaction, transaction_date)
                        )
                except Exception as e:
                    logger.error(f'There was an error on row { i + 1 } of {self.input_path}')
                    logger.error(transaction)
                    raise (e)

        self.transactions = sorted(transactions, key=itemgetter(FIELD_DATE))
        self._write_formatted_transactions()

    def _write_formatted_transactions(self):
        if self.output_path == '':
            raise ValueError(f'Cannot write formatted transactions, output_path missing')
        
        headers = list(self.transactions[0].keys())
        with open(self.output_path, 'w') as file_handler:
            writer = csv.DictWriter(file_handler, headers)
            writer.writeheader()
            for transaction in self.transactions:
                writer.writerow(transaction)

    def _rebase_transaction(self, transaction: dict, transaction_date: datetime) -> List[dict]:
        """ Splits a non-tax currency transaction into 2 transactions 
            denominated in the tax currency
        """
        # Many fields will be the same (e.g. date, note, id etc)
        base_transaction = dict(transaction)
        asset_transaction = dict(transaction)

        base_conversion_rate = self._get_conversion_rate(transaction_date, transaction[FIELD_BASE_CURRENCY])

        base_transaction[FIELD_QTY] = Decimal(transaction[FIELD_QTY]) * Decimal(transaction[FIELD_PRICE])
        base_transaction[FIELD_ASSET_CODE] = transaction[FIELD_BASE_CURRENCY]
        base_transaction[FIELD_PRICE] = base_conversion_rate

        asset_transaction[FIELD_QTY] = transaction[FIELD_QTY]
        asset_transaction[FIELD_ASSET_CODE] = transaction[FIELD_ASSET_CODE]
        asset_transaction[FIELD_PRICE] = Decimal(transaction[FIELD_PRICE]) * base_conversion_rate

        if transaction[FIELD_TRANSACTION_TYPE] == TRANSACTION_BUY_LABEL:
            # Buy asset, sell base
            # Example base = BTC asset = ETH, qty = 5 price = 0.025, type = buy
            base_transaction[FIELD_TRANSACTION_TYPE] = TRANSACTION_SELL_LABEL
            asset_transaction[FIELD_TRANSACTION_TYPE] = TRANSACTION_BUY_LABEL

            # Allocate the fee to the sell transaction only
            #   (rate conversion is handled in format_transaction if necessary)
            asset_transaction[FIELD_FEE] = '0'
            base_transaction[FIELD_FEE] = transaction[FIELD_FEE]

        else:
            # Sell asset, buy base
            # Example base = BTC asset = ETH, qty = 5 price = 0.025, type = sell
            base_transaction[FIELD_TRANSACTION_TYPE] = TRANSACTION_BUY_LABEL
            asset_transaction[FIELD_TRANSACTION_TYPE] = TRANSACTION_SELL_LABEL

            # Allocate the fee to the sell transaction only
            #   (rate conversion is handled in format_transaction if necessary)
            base_transaction[FIELD_FEE] = '0'
            asset_transaction[FIELD_FEE] = transaction[FIELD_FEE]

        return [
            self._format_transaction(base_transaction, transaction_date),
            self._format_transaction(asset_transaction, transaction_date),
        ]

    def _rebase_fee(self, transaction: dict, transaction_date: datetime) -> Decimal:
        """ Convert fee if not in base currency
        """
        if not transaction[FIELD_FEE] or transaction[FIELD_FEE] == '0':
            return Decimal(0)

        if self.tax_currency == transaction[FIELD_FEE_CURRENCY]:
            return Decimal(transaction[FIELD_FEE])

        conversion_rate = self._get_conversion_rate(transaction_date, transaction[FIELD_FEE_CURRENCY])
        return conversion_rate * Decimal(transaction[FIELD_FEE])

    def _get_conversion_rate(self, transaction_date: datetime, asset_code: str) -> Decimal:
        if self.rates is None:
            raise InputValidationError(f'No exchange rates provided')

        date_key = transaction_date.strftime(DATE_RATE_FORMAT)
        if self.rates.get(date_key) is None or self.rates[date_key].get(asset_code) is None:
            raise InputValidationError(
                f'Conversion rate missing from rates file for {asset_code} on {date_key}'
            )

        return Decimal(self.rates[date_key][asset_code])


    def _validate_transaction(self, transaction: dict):
        if transaction[FIELD_TRANSACTION_TYPE] not in (TRANSACTION_BUY_LABEL, TRANSACTION_SELL_LABEL):
            raise InputValidationError(f'Unknown transaction type {transaction[FIELD_TRANSACTION_TYPE]} ')

    def _standardize_transaction(self, transaction: dict) -> dict:
        return {
            **transaction,
            FIELD_ASSET_CODE: transaction[FIELD_ASSET_CODE].lower(),
            FIELD_BASE_CURRENCY: transaction[FIELD_BASE_CURRENCY].lower(),
            FIELD_FEE_CURRENCY: transaction[FIELD_FEE_CURRENCY].lower(),
        }



""" Outputs CG events to CSV, and outputs each CG event's corresponding cost base(s) to a separate CSV.
"""
import csv
from pathlib import Path
from typing import List

from capitalg.contstants import (
    FIELD_BASE_CURRENCY,
    FIELD_EXCHANGE,
    FIELD_FEE,
    FIELD_CAPITAL_GAIN_LT,
    FIELD_CAPITAL_GAIN_ST,
    FIELD_CAPITAL_GAIN_TOTAL,
    FIELD_COST_BASE_AMOUNT,
    FIELD_COST_BASE_ID,
    FIELD_ASSET_CODE,
    FIELD_DATE,
    FIELD_FEE_CURRENCY,
    FIELD_FEE_UNIT,
    FIELD_NOTE,
    FIELD_PRICE,
    FIELD_QTY,
    FIELD_RAW_ID,
    FIELD_SALE_AMOUNT,
    FIELD_SALE_BROKERAGE,
    FIELD_SALE_ID,
    FIELD_TRANSACTION_TYPE,
    FIELD_TZ,
    OUTPUT_FIELDS_CGT_EVENTS,
    OUTPUT_FIELDS_COST_BASE,
    OUTPUT_FIELDS_UNALLOCATED_COST_BASE,
)


class Writer:

    def __init__(self, cgt_events_path: str, cost_base_path: str):
        self.cgt_events_handler = open(cgt_events_path, 'w')
        self.cgt_events_writer = csv.DictWriter(self.cgt_events_handler, OUTPUT_FIELDS_CGT_EVENTS)
        self.cgt_events_writer.writeheader()

        # Cost base handler is used as an audit trail
        # It provides a bridge between the cost base of cgt event, and the raw transactions
        self.cost_base_handler = open(cost_base_path, 'w')
        self.cost_base_writer = csv.DictWriter(self.cost_base_handler, OUTPUT_FIELDS_COST_BASE)
        self.cost_base_writer.writeheader()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup()

    def cleanup(self):
        self.cgt_events_handler.close()
        self.cost_base_handler.close()

    def write_cgt_event(self, cost_base: dict, sale: dict):
        self.cgt_events_writer.writerow({
            FIELD_DATE: sale[FIELD_DATE],
            FIELD_SALE_ID: sale[FIELD_RAW_ID],
            FIELD_EXCHANGE: sale[FIELD_EXCHANGE],
            FIELD_COST_BASE_ID: cost_base[FIELD_COST_BASE_ID],
            FIELD_ASSET_CODE: sale[FIELD_ASSET_CODE],
            FIELD_QTY: sale[FIELD_QTY],
            FIELD_PRICE: sale[FIELD_PRICE],
            FIELD_SALE_AMOUNT: sale[FIELD_QTY] * sale[FIELD_PRICE],
            FIELD_SALE_BROKERAGE: sale[FIELD_FEE],
            FIELD_COST_BASE_AMOUNT: cost_base[FIELD_COST_BASE_AMOUNT],
            FIELD_CAPITAL_GAIN_TOTAL: cost_base[FIELD_CAPITAL_GAIN_TOTAL],
            FIELD_CAPITAL_GAIN_LT: cost_base[FIELD_CAPITAL_GAIN_LT],
            FIELD_CAPITAL_GAIN_ST: cost_base[FIELD_CAPITAL_GAIN_TOTAL] - cost_base[FIELD_CAPITAL_GAIN_LT],
            FIELD_NOTE:sale[FIELD_NOTE],
        })

    def write_cost_base_transactions(self, cost_base_id: str, cost_transactions: List[dict]):
        for cost_transaction in cost_transactions:
            self.cost_base_writer.writerow({
                FIELD_COST_BASE_ID: cost_base_id,
                **cost_transaction,
            })

    @staticmethod
    def output_unallocted_cost_base_transactions(output_file: Path, queues: dict):
        with open(output_file, 'w') as f:
            writer = csv.DictWriter(f, OUTPUT_FIELDS_UNALLOCATED_COST_BASE)
            writer.writeheader()

            for asset, queue in queues.items():
                for transaction in queue.get_queue:
                    writer.writerow({
                        FIELD_RAW_ID: transaction[FIELD_RAW_ID],
                        FIELD_EXCHANGE: transaction[FIELD_EXCHANGE],
                        FIELD_DATE: transaction[FIELD_DATE],
                        FIELD_TZ: transaction[FIELD_TZ],
                        FIELD_TRANSACTION_TYPE: transaction[FIELD_TRANSACTION_TYPE],
                        FIELD_BASE_CURRENCY: transaction[FIELD_BASE_CURRENCY],
                        FIELD_ASSET_CODE: transaction[FIELD_ASSET_CODE],
                        FIELD_QTY: transaction[FIELD_QTY],
                        FIELD_PRICE: transaction[FIELD_PRICE],
                        FIELD_FEE_CURRENCY: transaction[FIELD_FEE_CURRENCY],
                        FIELD_FEE: transaction[FIELD_FEE],
                        FIELD_NOTE: transaction[FIELD_NOTE],
                    })



import csv
from decimal import Decimal
from pathlib import Path

from ..contstants import (
    FIELD_ASSET_CODE,
    FIELD_FEE,
    FIELD_FEE_CURRENCY,
    FIELD_QTY,
    FILE_DIR,
    FILE_TRANSACTIONS,
    FILE_UNALLOCATED_COST_BASE_TRANSACTION,
)


def total_unallocated_cost_base(cost_base_file: Path) -> dict:
    assets = {}
    with open(cost_base_file) as f:
        reader = csv.DictReader(f)
        for _, row in enumerate(reader):
            if row[FIELD_ASSET_CODE] not in assets:
                assets[row[FIELD_ASSET_CODE]] = 0
            assets[row[FIELD_ASSET_CODE]] += Decimal(row[FIELD_QTY])
    return assets


def total_fees(tax_currency: str, transactions_file_path: Path) -> dict:
    assets = {}
    with open(transactions_file_path) as f:
        reader = csv.DictReader(f)
        for _, row in enumerate(reader):

            if row[FIELD_FEE_CURRENCY] == tax_currency:
                continue

            if row[FIELD_FEE_CURRENCY] not in assets:
                assets[row[FIELD_FEE_CURRENCY]] = 0
            assets[row[FIELD_FEE_CURRENCY]] += Decimal(row[FIELD_FEE])
    return assets


def get_balance(tax_currency: str, transactions_file_path: Path,
                cost_base_file: Path):
    unallocated_cost_base = total_unallocated_cost_base(cost_base_file)
    fees = total_fees(tax_currency, transactions_file_path)
    # NOTE this does not factor it withdrawal fees.
    #   To include, add each withdrawal as a separate sell transaction for price=0
    #   These will automatically draw from the cost base
    return {
        asset: balance - fees.get(asset, 0)
        for asset, balance in unallocated_cost_base.items()
    }


if __name__ == '__main__':
    balances = get_balance('aud', FILE_DIR / FILE_TRANSACTIONS,
                           FILE_DIR / FILE_UNALLOCATED_COST_BASE_TRANSACTION)
    print(balances)



import uuid
from datetime import datetime
from decimal import Decimal
from typing import Dict, List

from capitalg.contstants import (
    DATE_INPUT_FORMAT,
    FIELD_CAPITAL_GAIN_LT,
    FIELD_CAPITAL_GAIN_TOTAL,
    FIELD_COST_BASE_AMOUNT,
    FIELD_COST_BASE_ID,
    FIELD_DATE,
    FIELD_FEE_UNIT,
    FIELD_FEE,
    FIELD_PRICE,
    FIELD_QTY,
)
from capitalg.Writer import Writer


def register_cg_event(writer: Writer, sale: dict, costs: List[dict]):
    cost_base = calculate_cg_transaction(sale, costs)
    validate_cgt_qty(sale['qty'], cost_base['qty'], sale, costs)

    cost_base_id = make_cost_base_id()
    writer.write_cost_base_transactions(cost_base_id, costs)
    writer.write_cgt_event(
        cost_base={
            FIELD_COST_BASE_ID: cost_base_id,
            **cost_base,
        },
        sale=sale
    )


def calculate_cg_transaction(sale: dict, costs: List[dict]) -> Dict[str, Decimal]:
    """ Reduce list of purchase transactions to a total qty and purchase amount
    Brokerage is portioned based on qty.
    Brokerage is always included. If you don't want brokerage, 0 it out before calling this method

    sale_date must be in format YYYY-MM-DDTHH:MM:SS
    cost_transactions must contain keys FIELD_DATE in format YYYY-MM-DDTHH:MM:SS

    """
    qty = 0
    cost_base_amount = 0
    cap_gain_lt = 0

    for cost in costs:
        transaction_cost_base = calculate_cost_amount(cost)
        proceeds_from_qty = cost[FIELD_QTY] * (sale[FIELD_PRICE] - sale[FIELD_FEE_UNIT])
        transaction_cap_gain = proceeds_from_qty - transaction_cost_base
        qty += cost[FIELD_QTY]
        cost_base_amount += transaction_cost_base

        if is_long_term(sale[FIELD_DATE], cost[FIELD_DATE]) == True:
            cap_gain_lt += transaction_cap_gain

    total_cost_base_amount = cost_base_amount + sale[FIELD_FEE]

    return {
        FIELD_QTY: qty,
        FIELD_COST_BASE_AMOUNT: total_cost_base_amount,
        FIELD_CAPITAL_GAIN_TOTAL: qty * sale[FIELD_PRICE] - total_cost_base_amount,
        FIELD_CAPITAL_GAIN_LT: cap_gain_lt,
    }


def calculate_cost_amount(transaction: dict) -> Decimal:
    return transaction[FIELD_QTY] * (transaction[FIELD_PRICE] + transaction[FIELD_FEE_UNIT])


def validate_cgt_qty(sell_qty: Decimal, cost_qty: Decimal,
                     sell_transaction: dict, cost_transactions: List[dict]):
    if sell_qty > cost_qty:
        raise Exception(
            f'qty sold exceeds qty purchased. You may be missing cost transactions. Sell transaction: {sell_transaction}, cost transactions" {cost_transactions}'
        )

    if sell_qty < cost_qty:
        raise Exception(
            f'qty sold less than qty purchased from queue. There may be an error with CostBaseQueue. Sell transaction: {sell_transaction}, cost transactions" {cost_transactions}'
        )


def make_cost_base_id() -> str:
    return str(uuid.uuid4())


def is_long_term(sale_date_raw: str,
                 cost_date_raw: str,
                 date_format: str = DATE_INPUT_FORMAT) -> bool:
    sale_date = datetime.strptime(sale_date_raw, date_format)
    cost_date = datetime.strptime(cost_date_raw, date_format)
    return (sale_date - cost_date).days >= 365



import os
from pathlib import Path

DATE_INPUT_FORMAT = "%Y-%m-%dT%H:%M:%S"
DATE_RATE_FORMAT = '%Y-%m-%d'
TAX_YEAR_INPUT_FORMAT = '%Y-%m-%d'

FIELD_AMOUNT = 'amount'
FIELD_BASE_CURRENCY = 'base_currency'
FIELD_FEE = 'fee'
FIELD_FEE_CURRENCY = 'fee_currency'
FIELD_FEE_UNIT = 'fee_unit'
FIELD_CAPITAL_GAIN_LT = 'captial_gain_lt'
FIELD_CAPITAL_GAIN_ST = 'captial_gain_st'
FIELD_CAPITAL_GAIN_TOTAL = 'captial_gain_total'
FIELD_COST_BASE_AMOUNT = 'cost_base_amount'
FIELD_COST_BASE_ID = 'cost_base_id'
FIELD_ASSET_CODE = 'asset_code'
FIELD_DATE = 'date'
FIELD_EXCHANGE = 'exchange'
FIELD_NOTE = 'note'
FIELD_PRICE = 'price'
FIELD_QTY = 'qty'
FIELD_RAW_ID = 'id'
FIELD_SALE_AMOUNT = 'sale_amount'
FIELD_SALE_BROKERAGE = 'sale_brokerage'
FIELD_SALE_ID = 'sale_id'
FIELD_TRANSACTION_TYPE = 'type'
FIELD_TZ = 'tz'

OUTPUT_FIELDS_CGT_EVENTS = [
    FIELD_ASSET_CODE,
    FIELD_CAPITAL_GAIN_LT,
    FIELD_CAPITAL_GAIN_ST,
    FIELD_CAPITAL_GAIN_TOTAL,
    FIELD_COST_BASE_AMOUNT,
    FIELD_COST_BASE_ID,
    FIELD_DATE,
    FIELD_EXCHANGE,
    FIELD_NOTE,
    FIELD_PRICE,
    FIELD_QTY,
    FIELD_SALE_AMOUNT,
    FIELD_SALE_BROKERAGE,
    FIELD_SALE_ID,
]

OUTPUT_FIELDS_COST_BASE = [
    FIELD_AMOUNT,
    FIELD_ASSET_CODE,
    FIELD_BASE_CURRENCY,
    FIELD_COST_BASE_ID,
    FIELD_DATE,
    FIELD_EXCHANGE,
    FIELD_FEE_UNIT,
    FIELD_FEE,
    FIELD_FEE_CURRENCY,
    FIELD_NOTE,
    FIELD_PRICE,
    FIELD_QTY,
    FIELD_RAW_ID,
    FIELD_TRANSACTION_TYPE,
    FIELD_TZ,
]

OUTPUT_FIELDS_UNALLOCATED_COST_BASE = [
    FIELD_RAW_ID,
    FIELD_EXCHANGE,
    FIELD_DATE,
    FIELD_TZ,
    FIELD_TRANSACTION_TYPE,
    FIELD_BASE_CURRENCY,
    FIELD_ASSET_CODE,
    FIELD_QTY,
    FIELD_PRICE,
    FIELD_FEE_CURRENCY,
    FIELD_FEE,
    FIELD_NOTE,
]

# FIELD_TRANSACTION_TYPE must be one of these
TRANSACTION_BUY_LABEL = 'buy'
TRANSACTION_SELL_LABEL = 'sell'

FILE_DIR = Path(os.environ.get('FILE_DIR', './capitalg_files'))
FILE_TRANSACTIONS = 'transactions.csv'
FILE_FORMATTED_TRANSACTIONS = 'formatted_transactions.csv'
FILE_CG_EVENTS = 'cg_events.csv'
FILE_COST_BASE_TRANSACTION = 'cost_base_transactions.csv'
FILE_RATES = 'rates.csv'
FILE_UNALLOCATED_COST_BASE_TRANSACTION = 'unallocated_cost_base_transactions.csv'



import csv
from datetime import datetime, timedelta
from decimal import Decimal

from capitalg.contstants import (
    DATE_INPUT_FORMAT,
    FIELD_ASSET_CODE,
    FIELD_CAPITAL_GAIN_LT,
    FIELD_CAPITAL_GAIN_ST,
    FIELD_CAPITAL_GAIN_TOTAL,
    TAX_YEAR_INPUT_FORMAT,
)


def cg_summary(cg_events_path: str, tax_year_end: str) -> dict:

    tax_year_cutoff_obj = datetime.strptime(tax_year_end, TAX_YEAR_INPUT_FORMAT) + timedelta(days=1)
    tax_year_start_obj = datetime(year=tax_year_cutoff_obj.year - 1, month=tax_year_cutoff_obj.month, day=tax_year_cutoff_obj.day)

    cg_template = {
        FIELD_CAPITAL_GAIN_TOTAL: 0,
        FIELD_CAPITAL_GAIN_LT: 0,
        FIELD_CAPITAL_GAIN_ST: 0,
    }

    assets = {}
    with open(cg_events_path, 'r') as f:
        reader = csv.DictReader(f)
        for _, row in enumerate(reader):
            cg_date = datetime.strptime(row['date'], DATE_INPUT_FORMAT)
            if cg_date < tax_year_start_obj or cg_date >= tax_year_cutoff_obj:
                continue

            if row[FIELD_ASSET_CODE] not in assets:
                assets[row[FIELD_ASSET_CODE]] = cg_template.copy()

            assets[row[FIELD_ASSET_CODE]][FIELD_CAPITAL_GAIN_ST] += Decimal(row[FIELD_CAPITAL_GAIN_ST])
            assets[row[FIELD_ASSET_CODE]][FIELD_CAPITAL_GAIN_LT] += Decimal(row[FIELD_CAPITAL_GAIN_LT])
            assets[row[FIELD_ASSET_CODE]][FIELD_CAPITAL_GAIN_TOTAL] += Decimal(row[FIELD_CAPITAL_GAIN_TOTAL])

    total = cg_template.copy()
    for asset, cg in assets.items():
        total[FIELD_CAPITAL_GAIN_ST] += cg[FIELD_CAPITAL_GAIN_ST]
        total[FIELD_CAPITAL_GAIN_LT] += cg[FIELD_CAPITAL_GAIN_LT]
        total[FIELD_CAPITAL_GAIN_TOTAL] += cg[FIELD_CAPITAL_GAIN_TOTAL]

    assets.update(total=total)
    return assets



from pathlib import Path
from typing import List

from capitalg.contstants import (
    FIELD_ASSET_CODE,
    FIELD_QTY,
    FIELD_TRANSACTION_TYPE,
    TRANSACTION_BUY_LABEL,
    TRANSACTION_SELL_LABEL,
    FILE_TRANSACTIONS,
    FILE_CG_EVENTS,
    FILE_COST_BASE_TRANSACTION,
    FILE_FORMATTED_TRANSACTIONS,
    FILE_RATES,
    FILE_UNALLOCATED_COST_BASE_TRANSACTION,
)
from capitalg.CostBaseQueue import CostBaseQueue, QueueTypes
from capitalg.TransactionLoader import TransactionLoader
from capitalg.Writer import Writer
from capitalg.cg_helpers import register_cg_event
from capitalg.rates_loader import load_rates
from capitalg.utils import get_tax_year_cutoff_date


def calculate_cg(file_dir: Path, tax_currency: str, queue_type_code: str, tax_timezone: str, tax_year_end: str):

    loader = TransactionLoader(
        input_path=file_dir / FILE_TRANSACTIONS,
        output_path=file_dir / FILE_FORMATTED_TRANSACTIONS,
        tax_currency=tax_currency,
        tax_year_cutoff=get_tax_year_cutoff_date(tax_year_end, tax_timezone),
        tax_timezone=tax_timezone,
        rates=load_rates(file_dir / FILE_RATES)
    )

    process_transactions(
        file_dir=file_dir,
        transactions=loader.transactions,
        queue_type=QueueTypes.FIFO if queue_type_code == 'fifo' else QueueTypes.LIFO
    )


def process_transactions(file_dir: Path, transactions: List[dict], queue_type: QueueTypes):
    queues = {}

    with Writer(cgt_events_path=file_dir / FILE_CG_EVENTS, cost_base_path=file_dir / FILE_COST_BASE_TRANSACTION) as writer:

        for _, transaction in enumerate(transactions):

            asset_code = transaction[FIELD_ASSET_CODE]
            if asset_code not in queues:
                queues[asset_code] = CostBaseQueue(queue_type)

            if transaction[FIELD_TRANSACTION_TYPE] == TRANSACTION_BUY_LABEL:
                queues[asset_code].add(transaction)

            elif transaction[FIELD_TRANSACTION_TYPE] == TRANSACTION_SELL_LABEL:
                costs = queues[asset_code].get_transactions(transaction[FIELD_QTY])
                register_cg_event(writer=writer, costs=costs, sale=transaction)


        # Record unfulfilled cost base transactions
        # This will allow us to estimate our current asset balance
        writer.output_unallocted_cost_base_transactions(file_dir / FILE_UNALLOCATED_COST_BASE_TRANSACTION, queues)




from datetime import datetime, timedelta

import pytz

from capitalg.contstants import TAX_YEAR_INPUT_FORMAT


def convert_timezone(date_str: str, source_tz: str, dest_tz: str) -> datetime:
    dt = datetime.fromisoformat(date_str)
    dt_source_tz = pytz.timezone(source_tz).localize(dt)
    return dt_source_tz.astimezone(pytz.timezone(dest_tz))

def get_tax_year_cutoff_date(tax_year_end: str, tz: str, tax_year_end_format: str = TAX_YEAR_INPUT_FORMAT) -> datetime:
    """ Parse and localize the tax year's end date.
    The returned date object will be the start of the following tax year.
    This allows for easier date comparisons: All relevant transactions are less than this this cutoff date.
    """
    dt = datetime.strptime(tax_year_end, tax_year_end_format) + timedelta(days=1)
    return pytz.timezone(tz).localize(dt)



import argparse
from pathlib import Path

from capitalg.main import calculate_cg
from capitalg.contstants import FILE_DIR, FILE_CG_EVENTS, FILE_UNALLOCATED_COST_BASE_TRANSACTION, FILE_TRANSACTIONS
from capitalg.analysis.balance import get_balance
from capitalg.analysis.summary import cg_summary
from capitalg.contstants import FILE_DIR

def main():
    parser = argparse.ArgumentParser(prog='capitalg')
    subparsers = parser.add_subparsers()

    cg_parser = subparsers.add_parser('calculate', prog='capitalg calculate', description='Calculate captial gains from transactions.csv and rates.csv files')
    cg_parser.add_argument('-t', '--timezone', default='UTC', type=str, required=False, help='Tax reporting timezone. Valid timezones: https://gist.github.com/heyalexej/8bf688fd67d7199be4a1682b3eec7568')
    cg_parser.add_argument('-q', '--queue_type', choices=['fifo', 'lifo'], help='CG accounting method', required=True, type=str)
    cg_parser.add_argument('-c', '--tax_currency', help='Currency in which tax is to be paid', required=True, type=str.lower)
    cg_parser.add_argument('-d', '--tax_year_end', help='The last day of the tax year, in YYYY-MM-DD format. Use the latest available compplete tax year. Capital gains will be calculated for all prior years as well', required=True, type=str)
    cg_parser.add_argument('-p', '--folder_path', help='Path to folder containing input and output files', default=FILE_DIR, type=str)
    cg_parser.set_defaults(func=cg)

    sumamry_subparser = subparsers.add_parser('summary', prog='capitalg summary', description='Print a summary of capital gains for a given tax year. "capitalg calculate" must be run beforehand')
    sumamry_subparser.add_argument('-d', '--tax_year_end', help='The last day of the tax year, in YYYY-MM-DD format', required=True, type=str)
    sumamry_subparser.add_argument('-p', '--folder_path', help='Path to folder containing output files', default=FILE_DIR, type=str)
    sumamry_subparser.set_defaults(func=summary)

    balance_subparser = subparsers.add_parser('balance', prog='capitalg balance', description='Print an approximate asset balance. "capitalg calculate" must be run beforehand. The balance will be as at the tax_year_end used in the calculate command.')
    balance_subparser.add_argument('-c', '--tax_currency', help='Currency in which tax is to be paid', required=True, type=str.lower)
    balance_subparser.add_argument('-p', '--folder_path', help='Path to folder containing output files', default=FILE_DIR, type=str)
    balance_subparser.set_defaults(func=balance)

    args = parser.parse_args()
    args.func(args)


def cg(args):
    print('Calcualting capital gains...')
    calculate_cg(Path(args.folder_path), args.tax_currency, args.queue_type, args.timezone, args.tax_year_end)
    print(f'Finished calculating capital gains. Output files are available in the {args.folder_path} folder')

def summary(args):
    assets = cg_summary(Path(args.folder_path) / FILE_CG_EVENTS, args.tax_year_end)
    for asset, cgs in assets.items():
        print(asset)
        for k, v in cgs.items():
            print(f"{k}: { '{:,}'.format(round(v, 0)) }")
        print()

def balance(args):
    balances = get_balance(args.tax_currency, Path(args.folder_path) / FILE_TRANSACTIONS, Path(args.folder_path) / FILE_UNALLOCATED_COST_BASE_TRANSACTION)
    for asset, balance in balances.items():
        print(f"{asset}: { '{:,}'.format(round(balance, 4)) }")
    print()

if __name__ == '__main__':
    main()


import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setuptools.setup(
    name="capitalg",
    version="0.0.1",

    author="dleber",
    author_email="",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    description="Capital Gains Calculator",
    entry_points={
        "console_scripts": [
            "capitalg=capitalg.commands:main"
        ],
    },
    install_requires=[
       "pytz >= 2020.1",
    ],
    keywords =["capitalg", "capital gains", "capital gains tax calculator", "crypto tax",],
    license = "GNU",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["tests"]),
    project_urls={
        "Bug Tracker": "https://github.com/dleber/capitalg/issues",
    },
    # https://docs.python.org/3/distutils/setupscript.html#listing-whole-packages
    python_requires=">=3.6",
    url="https://github.com/dleber/capitalg",
)


import enum
from decimal import Decimal
from typing import Dict

from capitalg.contstants import FIELD_QTY


class QueueTypes(enum.Enum):
    LIFO = 'LIFO'
    FIFO = 'FIFO'


class CostBaseQueue:
    """ NOTE The queue does not validate transaction order
    The order of the queue is simply the order in which .add is called
    """

    def __init__(self, queue_type: QueueTypes):
        self.type = queue_type
        self.queue = []

    def add(self, row: Dict[str, Decimal]):
        """ Add a transaction for asset purchase
        """
        self.queue.append(row)

    def get_transactions(self, sale_qty: Decimal):
        transactions = []
        consumed_rows = 0
        qty = sale_qty

        queue = reversed(
            self.queue) if self.type == QueueTypes.LIFO else self.queue

        for i, row in enumerate(queue):

            if qty < row[FIELD_QTY]:
                copied_row = row.copy()
                copied_row[FIELD_QTY] = qty
                row[FIELD_QTY] -= qty
                transactions.append(copied_row)
                break

            elif qty >= row[FIELD_QTY]:
                copied_row = row.copy()
                consumed_rows += 1
                qty -= copied_row[FIELD_QTY]
                transactions.append(copied_row)
                continue

        # Remove fully sold purchases from queue
        if consumed_rows > 0:
            if self.type == QueueTypes.LIFO:
                del self.queue[-consumed_rows:]
            else:
                del self.queue[:consumed_rows]

        return transactions

    @property
    def get_queue(self):
        return self.queue
